/// @file			audio_record.c
/// @copyright  	Copyright (c) 2014 Toradex AG.
/// $Author:		khv
/// $Date: 			$ 2014-01-10 $
/// @brief:			File contains definitions for audio_record.h. 

/// Header files.
#include <windows.h>
#include "audio_record.h"

/// Allocates memory dynamically (Tries up to minimum of 1/4th of memory to allocate in case of low memory available).
BOOL AllocateMemory(BYTE** ptrToBuffer, DWORD *totalBytes)
{
	DWORD totalBytesToAllocate = 0;
	DWORD loopCount = 0;
	
	*ptrToBuffer = NULL;
	totalBytesToAllocate = *totalBytes;
	*totalBytes = 0;
	for (loopCount = 0; loopCount <= 2; loopCount++)
	{
		*ptrToBuffer = (BYTE*)malloc(totalBytesToAllocate);
		if (*ptrToBuffer != NULL)
		{
			break;
		}
		else
		{
			totalBytesToAllocate /= 2;
		}
	}
	if (*ptrToBuffer == NULL)
	{
		return FALSE;
	}
	*totalBytes = totalBytesToAllocate;
	return TRUE;
}

/// Starts recording.
BOOL StartRecording(HWAVEIN waveInHandle, 
					WAVEHDR *waveInHeaderPtr, 
					DWORD sizeOfWaveHeader, 
					CHAR *recordBuffer, 
					DWORD totalBytesToRecord)
{
	MMRESULT retVal = MMSYSERR_NOERROR;
	/// Un-prepares the header, if prepared for input.
	if ((waveInHeaderPtr->dwFlags & WHDR_PREPARED) || waveInHeaderPtr->reserved)
	{
		waveInUnprepareHeader(waveInHandle, waveInHeaderPtr, sizeOfWaveHeader);
	}
	/// Clears buffer before recording.
	memset((VOID*)recordBuffer, 0, totalBytesToRecord);
	memset((VOID*)waveInHeaderPtr, 0, sizeOfWaveHeader);
	/// Prepares waveheader for input.
	waveInHeaderPtr->dwBufferLength = totalBytesToRecord;
	waveInHeaderPtr->lpData = recordBuffer;
	/// Prepares a buffer for waveform input.
	retVal = waveInPrepareHeader(waveInHandle, waveInHeaderPtr, sizeOfWaveHeader);
	if (retVal != MMSYSERR_NOERROR)
	{
		return FALSE;		
	}
	/// Sends buffer to the waveform input device.
	retVal = waveInAddBuffer(waveInHandle, waveInHeaderPtr, sizeOfWaveHeader);
	if (retVal != MMSYSERR_NOERROR)
	{
		return FALSE;
	}
	/// Starts input on the waveform input device.
	retVal = waveInStart(waveInHandle);
	if (retVal != MMSYSERR_NOERROR)
	{
		return FALSE;
	}
	return TRUE;
}

/// Opens file to store the recording.
BOOL OpenFileToWrite(WCHAR *fileName, size_t fileNameLength, HANDLE *openFileHandle)
{
	BOOL success = FALSE;
	DWORD loopTerminate = FALSE;
	DWORD fileCount = 0;

	/// Creates directory in FlashDisk.
	success = CreateDirectory(L"\\FlashDisk\\audio_waves", NULL);
	if (success == FALSE)
	{
		if (GetLastError() != ERROR_ALREADY_EXISTS)
		{
			return FALSE;
		}
	}

	/// Creates file.
	while (loopTerminate == FALSE)
	{
		swprintf_s(fileName, fileNameLength, L"\\FlashDisk\\audio_waves\\Audio_wave_%d.wav", fileCount);
		*openFileHandle = CreateFile(fileName, GENERIC_WRITE, 0, 0, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
		if (*openFileHandle == NULL)
		{
			return FALSE;
		}
		/// If file exists before, close the file and scan further.
		if (GetLastError() == ERROR_ALREADY_EXISTS)
		{
			CloseHandle(*openFileHandle);
			fileCount++;
			continue;
		}
		loopTerminate = TRUE;
	}
	return TRUE;
}

/// Write file headers, format, and recorded data.
BOOL WriteFormatAndData(HANDLE fileHandle, 
						WAVEFORMATEX *ptrFormatChunk, 
						DWORD sizeOfFormatChunk, 
						CONST CHAR *recordBuffer, 
						DWORD totalBytesRecorded)
{
	riffFileHeader fileHeader = {0};
	riffFormatChunkHeader formatHeader = {0};
	riffDataChunkHeader dataHeader = {0};
	BOOL retVal = FALSE;
	DWORD bytesWritten = 0;

	/// Prepares format chunk header.
	formatHeader.formatChunkHeader = FMT_CHUNK_ID;
	formatHeader.formatChunkSize = sizeOfFormatChunk + ptrFormatChunk->cbSize;
	/// Prepares data chunk header.
	dataHeader.dataChunkHeader = DATA_CHUNK_ID;
	dataHeader.dataChunkSize = totalBytesRecorded;
	/// Prepares file header.
	fileHeader.riffHeader = RIFF_ID;
	fileHeader.riffHeaderSize = sizeof(fileHeader) + 
								(sizeof(formatHeader) + formatHeader.formatChunkSize) + 
								(sizeof(dataHeader) + dataHeader.dataChunkHeader);
	fileHeader.waveType = WAVE_ID;

	/// Write riffFileHeader.
	retVal = WriteFile(fileHandle, (VOID*)&fileHeader, sizeof(fileHeader), &bytesWritten, NULL);
	if (retVal == FALSE || (sizeof(fileHeader) != bytesWritten))
	{
		return FALSE;
	}
	/// Write riffFormatChunkHeader.
	retVal = WriteFile(fileHandle, (VOID*)&formatHeader, sizeof(formatHeader), &bytesWritten, NULL);
	if (retVal == FALSE || (sizeof(formatHeader) != bytesWritten))
	{
		return FALSE;
	}
	/// Write format chunk.
	retVal = WriteFile(fileHandle, (VOID*)ptrFormatChunk, sizeOfFormatChunk, &bytesWritten, NULL);
	if (retVal == FALSE || (sizeOfFormatChunk != bytesWritten))
	{
		return FALSE;
	}
	/// Write riffDataChunkHeader.
	retVal = WriteFile(fileHandle, (VOID*)&dataHeader, sizeof(dataHeader), &bytesWritten, NULL);
	if (retVal == FALSE || (sizeof(dataHeader) != bytesWritten))
	{
		return FALSE;
	}
	/// Write recorded data.
	retVal = WriteFile(fileHandle, (VOID*)recordBuffer, totalBytesRecorded, &bytesWritten, NULL);
	if (retVal == FALSE || (totalBytesRecorded != bytesWritten))
	{
		return FALSE;
	}
	return TRUE;
}