/// @file			audio_record.h
/// @copyright  	Copyright (c) 2014 Toradex AG.
/// $Author:		khv
/// $Date: 			$ 2014-01-10 $
/// @brief:			File contains declarations to support audio recording. 

#ifndef _AUDIO_RECORD_H_
#define _AUDIO_RECORD_H_

/// Header files.
#include <windows.h>

/// Macros.
#define RIFF_ID					0x46464952		///< "RIFF" Little Endian format.
#define WAVE_ID					0x45564157		///< "WAVE" Little Endian format.
#define FMT_CHUNK_ID			0x20746d66		///< "fmt " Little Endian format.
#define DATA_CHUNK_ID			0x61746164		///< "data" Little Endian format.
#define ID_LENGTH						 4		///< Total bytes in Id.
#define DATA_BYTES_IN_FORMAT_CHUNK		16		///< Without extra format bytes.
#define MAX_VOL					0xFFFFFFFF		///< Maximum volume.

/// RIFF file header.
typedef struct tRiffFileHeader
{
	DWORD   riffHeader;
	DWORD   riffHeaderSize;
	DWORD   waveType;
}riffFileHeader;

/// RIFF format chunk header.
typedef struct tRiffFormatChunkHeader
{
	DWORD   formatChunkHeader;   
	DWORD   formatChunkSize;
}riffFormatChunkHeader;

/// RIFF data chunk header.
typedef struct tRiffDataChunkHeader
{
	DWORD   dataChunkHeader;   
	DWORD   dataChunkSize;
}riffDataChunkHeader;

/// Current state. 
typedef enum tState
{
	UNDEFINED,
	PAUSED,
	STOPPED,
	PLAYING,
	RECORDING,
	RECORDING_DONE
}state;

/// Operation to perform.
typedef enum tOperation
{
	RECORD = 1,
	FINISH_RECORD,
	SAVE,
	PLAY,
	PAUSE,
	STOP,
	QUIT
}operation;
#endif

//******************************************************************************
/// Allocates memory dynamically (Tries up to minimum of 1/4th of memory to allocate in case of low memory available).
/// @param[out]		buffer       Pointer to memory allocated buffer.
/// @param[inout]   totalBytes   In - Total bytes to allocate.
///								 Out - Total bytes allocated.
/// @retval			TRUE         Success.
///					FALSE        Failure.
BOOL AllocateMemory(BYTE** ptrToBuffer, DWORD *totalBytes);

//******************************************************************************
/// Starts recording on waveform input device.
/// @param[in]		waveInHandle	     Handle to open waveform input device.
/// @param[inout]	waveInHeaderPtr		 Header for waveform input device.
/// @param[out]		bufferRecord		 Buffer recorded.
/// @param[in]		totalbytesToRecord   Total bytes to record.
/// @retval			TRUE				 Success.
///					FALSE				 Failure.
BOOL StartRecording(HWAVEIN waveInHandle, 
					WAVEHDR *waveInHeaderPtr, 
					DWORD sizeOfWaveHeader, 
					CHAR *bufferRecord, 
					DWORD totalbytesToRecord);

//******************************************************************************
/// Opens file to store the recording.
/// @param[out]		fileName		  Name of the file where recording is to be stored.
/// @param[in]		fileNameLength    Size of the fileName buffer.
/// @param[in]		sizeOfWaveHeader  Size of wave header.
/// @param[out]		openFileHandle    Handle to the file where recording is to be stored. 
/// @retval			TRUE			  Success.
///					FALSE			  Failure.
BOOL OpenFileToWrite(WCHAR *fileName, size_t fileNameLength, HANDLE *openFileHandle);

//******************************************************************************
/// Write file headers, format, and recorded data.
/// @param[in]		fileHandle		     handle to the file where recording is to be stored.
/// @param[in]		ptrFormatChunk       Pointer to format chunk.
/// @param[in]		sizeOfFormatChunk    Size of format chunk.
/// @param[in]		recordBuffer         Buffer where recording is stored.
/// @param[in]		totalBytesRecorded	 Total number of bytes stored in recordBuffer.
/// @retval			TRUE			     Success.
///					FALSE			     Failure.
BOOL WriteFormatAndData(HANDLE fileHandle, 
						WAVEFORMATEX *ptrFormatChunk, 
						DWORD sizeOfFormatChunk, 
						CONST CHAR *recordBuffer, 
						DWORD totalBytesRecorded);