/// Header files.
#include <windows.h>
#include <mmsystem.h>
#include "audio_record.h"

/// Macros.
#define RECORDING_LENGTH_IN_SECOND 60

/// Procedures. 
/// Based on time-out, function states that recording is complete.
DWORD ShowTimeout(LPVOID);
/// Clean up.
void Cleanup(state currentState);

/// Static declarations.
static WAVEHDR waveInHeader = {0};
static WAVEHDR waveOutHeader = {0};
static HWAVEIN waveIn = {0};
static HWAVEOUT waveOut = {0};
static CHAR *recordAndPlayBuffer = NULL;
static HANDLE timerThread = NULL;

int wmain()
{
	/// Local declarations.
	WCHAR fileName[50] = {0};
	HANDLE fileHandle = NULL;
	WAVEFORMATEX formatChunk = {0};
	DWORD select = 0;
	state currentState = UNDEFINED;
	DWORD totalBytesRecorded = 0;
	DWORD totalBytesToAllocate = 0;
	MMRESULT retVal = MMSYSERR_NOERROR;
	BOOL success = FALSE;
	DWORD recordLengthInSec = RECORDING_LENGTH_IN_SECOND;

	/// Get number of waveform input devices present in the system.
	if (waveInGetNumDevs() == 0)
	{
		printf("\nNo waveform input device present in the system for recording.");
		fflush(stdin);
		getchar();
		return 0;
	}
	/// Get number of waveform output devices present in the system.
	if (waveOutGetNumDevs() == 0)
	{
		printf("\nNo waveform output device present in the system for playback.");
		fflush(stdin);
		getchar();
		return 0;
	}
	/// Set waveform format parameters.
	formatChunk.cbSize = 0;
	formatChunk.wBitsPerSample = 16;
	formatChunk.nSamplesPerSec = 44100;
	formatChunk.nChannels = 2;
	formatChunk.wFormatTag = WAVE_FORMAT_PCM;
	formatChunk.nBlockAlign = (formatChunk.wBitsPerSample * formatChunk.nChannels) / 8;
	formatChunk.nAvgBytesPerSec = formatChunk.nSamplesPerSec * formatChunk.nBlockAlign;
	/// Opens a waveform input device for recording purpose.
	retVal = waveInOpen(&waveIn, WAVE_MAPPER, &formatChunk, (DWORD_PTR)0, 0, CALLBACK_NULL);
	if (retVal != MMSYSERR_NOERROR)
	{
		printf("\nUnable to open the device for recording.");
		Cleanup(currentState);
		fflush(stdin);
		getchar();
		return 0;
	}
	/// Opens a waveform output device for playback.
	retVal = waveOutOpen(&waveOut, WAVE_MAPPER, &formatChunk, (DWORD_PTR)0, 0, CALLBACK_NULL);
	if (retVal != MMSYSERR_NOERROR)
	{
		printf("\nUnable to open the device for playback.");
		Cleanup(currentState);
		fflush(stdin);
		getchar();
		return 0;
	}

	while (select != QUIT)
	{
		printf("\n\n********************** Menu **********************");
		printf("\n1. Record \n2. Finish record \n3. Save to disk \n4. Play \n5. Pause \n6. Stop \n7. Quit\t");
		fflush(stdin);
		scanf_s("%u", &select);
		switch (select)
		{
		case RECORD:
			{
				if (currentState == RECORDING)
				{
					printf("\nAlready in record mode.");
					break;
				}
				else if (currentState == PLAYING || currentState == PAUSED)
				{
					printf("\nFirst stop the playback.");
					break;
				}
				/// Frees memory, if allocated.
				if (recordAndPlayBuffer != NULL)
				{
					free(recordAndPlayBuffer);
				}
				totalBytesToAllocate = formatChunk.nAvgBytesPerSec * recordLengthInSec;
				/// Allocates memory.
				if (!AllocateMemory((BYTE**)&recordAndPlayBuffer, &totalBytesToAllocate))
				{
					printf("\nMemory not enough. Program will terminate.");
					Cleanup(currentState);
					fflush(stdin);
					getchar();
					return 0;
				}
				/// Shows recording duration to the user.
				recordLengthInSec = totalBytesToAllocate / formatChunk.nAvgBytesPerSec;
				if (recordLengthInSec == RECORDING_LENGTH_IN_SECOND)
				{
					printf("\nTotal recording duration = %u seconds.", recordLengthInSec);
				}
				else
				{
					printf("\nMemory not enough. Recording duration reduced to %u seconds", recordLengthInSec);
				}
				timerThread = NULL;
				/// Creates a thread to tell the user for time-out of recording.
				timerThread = CreateThread(NULL, 0, &ShowTimeout, &recordLengthInSec, CREATE_SUSPENDED, NULL);
				/// Starts input on the waveform input device.
				success = StartRecording(waveIn, &waveInHeader, sizeof(waveInHeader), recordAndPlayBuffer, totalBytesToAllocate);
				if (success == FALSE)
				{
					printf("\nUnable to start recording.");
					select = QUIT;
					fflush(stdin);
					getchar();
					break;
				}
				else
				{
					ResumeThread(timerThread);
					printf("\nRecording...");
					currentState = RECORDING;
				}
				break;
			}
		case FINISH_RECORD:
			{
				if (currentState != RECORDING)
				{
					printf("\nFirst start the recording.");
					break;
				}
				/// Suspends the timer thread.
				SuspendThread(timerThread);
				/// Stops input waveform input device.
				waveInStop(waveIn);
				/// Stops input waveform input device and resets the current position to 0.
				waveInReset(waveIn);
				/// Close the timer thread handle.
				CloseHandle(timerThread);
				totalBytesRecorded = waveInHeader.dwBytesRecorded;
				/// Clears WHDR_DONE and WHDR_INQUEUE flags.
				waveInHeader.dwFlags &= 0x0E;
				printf("\nRecording done.");
				currentState = RECORDING_DONE;
				break;
			}
		case SAVE:
			{
				if (currentState == UNDEFINED)
				{
					printf("\nFirst record a file.");
					break;
				}
				else if (currentState == RECORDING)
				{
					printf("\nFirst stop recording.");
					break;
				}
				/// Opens handle to file to store recording data.
				success = OpenFileToWrite(&fileName, sizeof(fileName), &fileHandle);
				if (success == FALSE)
				{
					printf("\nUnable to create the file to store.");
					break;
				}
				printf("\nFile path:  %ls", fileName);
				printf("\nSaving... (Please wait)");
				/// Stores recording data to the file.
				success = WriteFormatAndData(fileHandle, &formatChunk, sizeof(formatChunk), recordAndPlayBuffer, totalBytesRecorded);
				if (success == FALSE)
				{
					if (GetLastError() == ERROR_DISK_FULL)
					{
						printf("\nMemory not enough. Save failed.");
					}
					else
					{
						printf("\nUnable to write the file.");
					}
					break;
				}
				printf("\nFile saved successfully.");
				/// Closes file handle.
				CloseHandle(fileHandle);
				break;
			}
		case PLAY:
			{
				if (currentState == UNDEFINED)
				{	
					printf("\nNo data to play.");
					break;
				}
				else if (currentState == RECORDING)
				{
					printf("\nFirst stop the recording.");
					break;
				}
				else if (currentState == RECORDING_DONE)
				{
					/// Un-prepares the header, if prepared for output.
					waveOutUnprepareHeader(waveOut, &waveOutHeader, sizeof(waveOutHeader));
					/// Prepares waveheader for output.
					waveOutHeader.dwBufferLength = totalBytesRecorded;
					waveOutHeader.lpData = recordAndPlayBuffer;
					/// Clears WHDR_DONE and WHDR_INQUEUE flags.
					waveOutHeader.dwFlags &= 0x0E;
					/// Prepares a waveform data block for playback.
					retVal = waveOutPrepareHeader(waveOut, &waveOutHeader, sizeof(waveOutHeader));
					if (retVal != MMSYSERR_NOERROR)
					{
						printf("\nUnable to prepare the data to play.");
						select = QUIT;
						fflush(stdin);
						getchar();
						break;
					}
				}
				else if (currentState == PLAYING && !(waveOutHeader.dwFlags & WHDR_DONE))
				{
					printf("\nAlready in play mode.");
					break;
				}
				if (currentState == PAUSED)
				{
					waveOutRestart(waveOut);
					printf("\nResuming...");
				}				
				else
				{
					waveOutHeader.dwFlags &= 0x0E;
					/// Sends a data block to the specified waveform output device.
					retVal = waveOutWrite(waveOut, &waveOutHeader, sizeof(waveOutHeader));
					if (retVal != MMSYSERR_NOERROR)
					{
						printf("\nUnable to play.");
						select = QUIT;
						fflush(stdin);
						getchar();
						break;
					}
					waveOutSetVolume(waveOut, MAX_VOL);
					printf("\nPlaying record...");
				}
				currentState = PLAYING;
				break;
			}
		case PAUSE:
			{
				if (currentState == PLAYING)
				{
					/// Pauses playback on a specified waveform output device.
					waveOutPause(waveOut);
					printf("\nPaused.");
					currentState = PAUSED;
				}
				else
				{
					printf("\nFirst start the playback.");
				}
				break;
			}
		case STOP:
			{
				if (currentState != PLAYING)
				{
					printf("\nFirst start the playback.");
				}
				else
				{
					/// Stops playback on a specified waveform output device.
					waveOutReset(waveOut);
					/// Clears WHDR_DONE and WHDR_INQUEUE flags.
					waveOutHeader.dwFlags &= 0x0E;
					currentState = STOPPED;
					printf("\nStopped.");
				}
				break;
			}
		case QUIT:
			{
				break;
			}
		default:
			{
				printf("\nInvalid selection. Try again.");
			}
		}
	}
	/// Clean up.
	Cleanup(currentState);
	return 0;
}

/// Based on time-out, function states that recording is complete.
DWORD ShowTimeout(LPVOID recordTime)
{
	DWORD recordDurationInSec = 0;
	recordDurationInSec = *((DWORD*)recordTime);
	while (recordDurationInSec > 0)
	{
		/// Wait for 1 second.
		Sleep(1000);
		recordDurationInSec -= 1;
	}
	printf("\n\nRecording timeout. Press 2 to \'Finish record\'.  ");
	return 0;
}

/// Clean up.
void Cleanup(state currentState)
{
	if (currentState == RECORDING)
	{
		/// Stops input waveform input device.
		waveInStop(waveIn);
		/// Stops input waveform input device and resets the current position to 0.
		waveInReset(waveIn);
	}
	else if (currentState == PLAYING)
	{
		/// Stops playback on a specified waveform output device.
		waveOutReset(waveOut);
	}
	/// Un-prepares the header, if prepared for input.
	if ((waveInHeader.dwFlags & WHDR_PREPARED) || waveInHeader.reserved)
	{
		waveInUnprepareHeader(waveIn, &waveInHeader, sizeof(waveInHeader));
	}
	/// Un-prepares the header, if prepared for output.
	if ((waveOutHeader.dwFlags & WHDR_PREPARED) || waveOutHeader.reserved)
	{
		waveOutUnprepareHeader(waveOut, &waveOutHeader, sizeof(waveOutHeader));
	}
	/// De-allocates the memory, if allocated.
	if (recordAndPlayBuffer != NULL)
	{
		free(recordAndPlayBuffer);
	}
	/// Closes the waveform-audio input device, if open.
	if (waveIn != NULL)
	{
		waveInClose(waveIn);
	}
	/// Closes the waveform-audio output device, if open.
	if (waveOut != NULL)
	{
		waveOutClose(waveOut);
	}
	/// Closes the timer thread handle, if open.
	if (timerThread != NULL)
	{
		CloseHandle(timerThread);
	}
}